--
-- FoldingParts
--
-- @author:    	Xentro (Marcus@Xentro.se)
-- @website:	www.Xentro.se
-- @history:	v0.7 - 2014-03-15 - First Build
-- 

FoldingParts = {};

function FoldingParts.prerequisitesPresent(specializations)
    return SpecializationUtil.hasSpecialization(AnimatedVehicle, specializations) and not SpecializationUtil.hasSpecialization(Foldable, specializations);
end;

function FoldingParts:load(xmlFile)
	self.setFold = SpecializationUtil.callSpecializationsFunction("setFold");
	self.getIsFoldDisabled = FoldingParts.getIsFoldDisabled;
	self.getToggledFoldDirection = FoldingParts.getToggledFoldDirection;
	self.getIsFoldAllowed = Utils.overwrittenFunction(self.getIsFoldAllowed, FoldingParts.getIsFoldAllowed);
	
	self.applyInitialAnimation = Utils.overwrittenFunction(self.applyInitialAnimation, FoldingParts.applyInitialAnimation);
	
	self.foldingParts = {};
	local i = 0;
	while true do
		local key = string.format("vehicle.FoldingParts.set(%d)", i);
		if not hasXMLProperty(xmlFile, key) then break; end;
		
		local foldInputButton = getXMLString(xmlFile, key .. "#foldInputButton");
		local animation = getXMLString(xmlFile, key .. "#foldAnimationName");
		
		if foldInputButton ~= nil then
			foldInputButton = InputBinding[foldInputButton];
		end;
		
		if animation ~= nil then
			if self.animations[animation] ~= nil then
				local entry = {};
				entry.animationName = animation;
				entry.foldMoveDirection = 1;
				entry.disabled = false;
				entry.posDirectionText = Utils.getNoNil(getXMLString(xmlFile, key .. "#posDirectionText"), "fold_OBJECT");
				entry.negDirectionText = Utils.getNoNil(getXMLString(xmlFile, key .. "#negDirectionText"), "unfold_OBJECT");
				entry.foldInputButton = Utils.getNoNil(foldInputButton, InputBinding.IMPLEMENT_EXTRA2);
				
				entry.foldAnimTime = 0;
				entry.maxFoldAnimDuration = math.max(0.0001, self:getAnimationDuration(entry.animationName));
				
				entry.startAnimTime = getXMLFloat(xmlFile, key .. "#startAnimTime");
				if entry.startAnimTime == nil then
					entry.startAnimTime = 0;
					local startMoveDirection = Utils.getNoNil(getXMLInt(xmlFile, key .. "#startMoveDirection"), 0);
					
					if startMoveDirection > 0.1 then
						entry.startAnimTime = 1;
					end;
				end;
				
				local parentSetIndice = getXMLInt(xmlFile, key .. "#parentSetIndice");
				if parentSetIndice ~= nil then
					entry.parentSetIndice = parentSetIndice + 1;
				end;
				
				table.insert(self.foldingParts, entry);
				i = i + 1;
			else
				print("FoldingParts - Error: Invalid foldAnimationName in " .. self.configFileName);
				break;
			end;
		else
			break;
		end;
	end;
	
	if table.getn(self.foldingParts) > 0 then
		self.isSelectable = true;
	end;
end;

function FoldingParts:applyInitialAnimation(superFunc)
	for part, v in ipairs(self.foldingParts) do
		v.foldAnimTime = 1 - v.startAnimTime;
		FoldingParts.setAnimTime(self, part, v.startAnimTime);
	end;
	
	if superFunc ~= nil then
		superFunc(self);
	end;
end;

function FoldingParts:setRelativePosition(positionX, offsetY, positionZ, yRot)
	for part, v in ipairs(self.foldingParts) do
		v.foldAnimTime = 1 - v.startAnimTime;
		FoldingParts.setAnimTime(self, part, v.startAnimTime);
	end;
end;
  
function FoldingParts:delete()
end;

function FoldingParts:readStream(streamId, connection)
	for part, v in ipairs(self.foldingParts) do
		local direction = streamReadInt8(streamId);
		local animTime = streamReadFloat32(streamId);
		FoldingParts.setAnimTime(self, part, animTime);
		self:setFold(part, direction, true);
	end;
end;

function FoldingParts:writeStream(streamId, connection)
	for _, v in ipairs(self.foldingParts) do
		local direction = v.foldMoveDirection;
		streamWriteInt8(streamId, direction);
		streamWriteFloat32(streamId, v.foldAnimTime);
	end;
end;

function FoldingParts:loadFromAttributesAndNodes(xmlFile, key, resetVehicles)
	if not resetVehicles then
		local i = 0;
		while true do
			local foldKey = key .. string.format(".foldingParts(%d)", i);
			if not hasXMLProperty(xmlFile, foldKey) then break; end;
			
			local part = self.foldingParts[i + 1];
			local animTime = getXMLFloat(xmlFile, foldKey .. "#animeTime");
			
			if animTime == nil then
				if part ~= nil then
					animTime = part.startAnimTime;
				end;
			end;
			
			if part == nil then
				print("Warning: Corrupt savegame, foldingParts " .. i + 1 .. " could not be loaded");
				break;
			end;
			
			local dir = -1;
			if animTime >= 0.5 then
				dir = 1;
			end;
			part.foldMoveDirection = dir;
			FoldingParts.setAnimTime(self, i + 1, animTime);
			i = i + 1;
		end;
	end;
	
	return BaseMission.VEHICLE_LOAD_OK;
end;

function FoldingParts:getSaveAttributesAndNodes(nodeIdent)
	local attributes = "";
	local nodes = "";
	
	for part, v in ipairs(self.foldingParts) do
		if part > 1 then
			nodes = nodes .. "\n";
		end;
		
		nodes = nodes .. nodeIdent .. '<foldingParts animeTime="' .. v.foldAnimTime .. '" />';
	end;
	
	return attributes, nodes;
end;

function FoldingParts:mouseEvent(posX, posY, isDown, isUp, button)
end;

function FoldingParts:keyEvent(unicode, sym, modifier, isDown)
end;

function FoldingParts:update(dt)
	if self:getIsActive() then
		for set, v in ipairs(self.foldingParts) do
			if math.abs(v.foldMoveDirection) > 0.1 then
				local foldAnimTime = 0;
				
				if v.foldMoveDirection < -0.1 then
					foldAnimTime = 1;
				end;
				  
				if v.foldMoveDirection > 0 then
					foldAnimTime = math.max(foldAnimTime, self:getRealAnimationTime(v.animationName) / v.maxFoldAnimDuration);
				elseif v.foldMoveDirection < 0 then
					foldAnimTime = math.min(foldAnimTime, self:getRealAnimationTime(v.animationName) / v.maxFoldAnimDuration);
				end;
				
				v.foldAnimTime = Utils.clamp(foldAnimTime, 0, 1);
			end;
			
			if not self:getIsFoldDisabled(set) then
				if self:getIsActiveForInput() then
					if InputBinding.hasEvent(v.foldInputButton) then
						if self:getIsFoldAllowed() then
							self:setFold(set, self:getToggledFoldDirection(set));
						end;
					end;
				end;
			end;
		end;
	end;
end;

function FoldingParts:updateTick(dt)
end;

function FoldingParts:draw()
	if self:getIsActiveForInput(true) then
		for i, v in ipairs(self.foldingParts) do
			if not self:getIsFoldDisabled(i) then
				if self:getIsFoldAllowed() then
					if self:getToggledFoldDirection(i) > 0 then
						g_currentMission:addHelpButtonText(string.format(g_i18n:getText(v.posDirectionText), self.typeDesc), v.foldInputButton);
					else
						g_currentMission:addHelpButtonText(string.format(g_i18n:getText(v.negDirectionText), self.typeDesc), v.foldInputButton);
					end;
				end;
			end;
		end;
	end;
end;

function FoldingParts:onDetach()
	if self.deactivateOnDetach then
		FoldingParts.onDeactivate(self);
	end;
end;

function FoldingParts:onDeactivate()
	for i, v in ipairs(self.foldingParts) do
		self:setFold(i, 0, true);
	end;
end;
  
function FoldingParts:setFold(partID, direction, noEventSend)
	local part = self.foldingParts[partID];
	
	if part ~= nil then
		if part.foldMoveDirection ~= direction then
			if noEventSend == nil or noEventSend == false then
				if g_server ~= nil then
					g_server:broadcastEvent(FoldingPartsSetFoldDirectionEvent:new(self, partID, direction), nil, nil, self);
				else
					g_client:getServerConnection():sendEvent(FoldingPartsSetFoldDirectionEvent:new(self, partID, direction));
				end;
			end;
			
			part.foldMoveDirection = direction;
			
			local speedScale = nil;
			if (direction > 0.1 or direction < -0.1) then
				speedScale = direction;
			end;
			
			self:stopAnimation(part.animationName, true);
			
			if speedScale ~= nil then
				local animTime = (part.foldAnimTime * part.maxFoldAnimDuration) / self:getAnimationDuration(part.animationName);
				self:playAnimation(part.animationName, speedScale, animTime, true);
			end;
		end;
	end;
end;

function FoldingParts:getIsFoldDisabled(partID)
	local part = self.foldingParts[partID];
	
	if part ~= nil then
		if part.parentSetIndice ~= nil then
			local parent = self.foldingParts[part.parentSetIndice];
			
			if not part.disabled and parent ~= nil then
				local targetTime = 1;
				
				if parent.startAnimTime > 0 then
					targetTime = 0;
				end;
				
				local time = self:getAnimationTime(parent.animationName)
				return (parent.startAnimTime > 0 and time > 0 or parent.startAnimTime == 0 and time < 1);
			else
				return true;
			end;
		else
			return part.disabled;
		end;
	end;
	
	return false;
end;

function FoldingParts:getIsFoldAllowed(superFunc)
	if superFunc ~= nil then
		return superFunc(self);
	end;
	
	return true;
end;

function FoldingParts:getToggledFoldDirection(partID)
	local part = self.foldingParts[partID];
	
	if part ~= nil then
		if (part.foldMoveDirection > 0.1 or part.foldMoveDirection == 0 and part.foldAnimTime >= (part.maxFoldAnimDuration / 0.5)) then
			return -1;
		else
			return 1;
		end;
	end;
end;

function FoldingParts.setAnimTime(self, partID, animTime)
	local part = self.foldingParts[partID];
	
	local currentTime = part.foldAnimTime;
	part.foldAnimTime = animTime;
	
	local speed = 1;
	if part.foldAnimTime < currentTime then
		speed = -1;
	end;
	
	local stopAnimTime = (part.foldAnimTime * part.maxFoldAnimDuration) / self:getAnimationDuration(part.animationName);
	self:playAnimation(part.animationName, speed, nil, true);
	self:setAnimationStopTime(part.animationName, stopAnimTime);
	AnimatedVehicle.updateAnimationByName(self, part.animationName, 99999999);

	if self.updateCylinderedInitial ~= nil then
		self:updateCylinderedInitial();
	end;
end;


FoldingPartsSetFoldDirectionEvent = {};
FoldingPartsSetFoldDirectionEvent_mt = Class(FoldingPartsSetFoldDirectionEvent, Event);

InitEventClass(FoldingPartsSetFoldDirectionEvent, "FoldingPartsSetFoldDirectionEvent");

function FoldingPartsSetFoldDirectionEvent:emptyNew()
	local self = Event:new(FoldingPartsSetFoldDirectionEvent_mt);
	
	return self;
end;

function FoldingPartsSetFoldDirectionEvent:new(object, part, direction)
	local self = FoldingPartsSetFoldDirectionEvent:emptyNew()
	self.object = object;
	self.part = part;
	self.direction = direction;
	
	return self;
end;

function FoldingPartsSetFoldDirectionEvent:readStream(streamId, connection)
	local id = streamReadInt32(streamId);
	self.object = networkGetObject(id);
	self.part = streamReadInt8(streamId);
	self.direction = streamReadInt8(streamId);
	
	self:run(connection);
end;

function FoldingPartsSetFoldDirectionEvent:writeStream(streamId, connection)
	streamWriteInt32(streamId, networkGetObjectId(self.object));
	streamWriteInt8(streamId, self.part);
	streamWriteInt8(streamId, self.direction);
end;

function FoldingPartsSetFoldDirectionEvent:run(connection)
	self.object:setFold(self.part, self.direction, true);
	
	if not connection:getIsServer() then
		g_server:broadcastEvent(FoldingPartsSetFoldDirectionEvent:new(self.object, self.part, self.direction), nil, connection, self.object);
	end;
end;